#!/usr/bin/perl -- -*- cperl -*-

#  This file is part of PuDaC.

#  PuDaC is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.

#  PuDaC is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with PuDaC  If not, see <http://www.gnu.org/licenses/>.

use Carp;
use Data::Dumper;
use Getopt::Std;

use strict;
use warnings;

# interpolation of $1 &c should be done by perl in the parser
# ... well, in objects. in messages it's messy. i'll have to figure that out.
# also, need to figure out exactly how I'm doing send/receive.
# and metro.

our %objectlist;
our ($globalinit, $loadbangs, $periteration, $dsptrace)=("","","","");
our $objectcount = 0;

our %objecttranslations;
our %xtypes;
our %parser;
our @unimplementedObjects;
do "pudac.objects.pl" or croak "Object translations broken";
do "pudac.xtypes.pl" or croak "Xtypes broken";
do "pudac.parser.pl" or croak "Parser broken";

my %opts;

getopts('o:h',\%opts);

croak "Use: pudac [-o {(default)alsa,oss,stdout}] infile.pd > outfile.c"
  if ($opts{h});

croak "OSS output not yet supported"
  if (lc $opts{o} eq "oss");

$opts{o} = "alsa"
  unless (defined $opts{o});

# Lexer time!

# (i.e. build data structures in memory corresponding to input .pd file)
my $FH;
open $FH, '<', $ARGV[0];

{
  while (my $line = <$FH>) {
    # need to fix this: it will not work with embedded escaped semicolons (\;)
    $line .= <$FH>
      until (not $line or $line =~ m.[^\\];.); # not escaped semicolons
    $line =~ s/;$//;

    # escape all " %
    # (don't need to escape \, they can't exist)

    $line =~ s/"/\"/g; # escape " (C)
    $line =~ s/%/\%/g; # escape % (printf)
    $line =~ s'\\\$'$'g; # unescape $
    $line =~ s/\\;/;/g; # unescape ;
    $line =~ s/\\,/,/g; # unescape ,

    my @parse = split /\s/,$line;
    # Should escape any quotation marks here!

    if ($#parse > 0) {
      #print STDERR join(":", @parse)."\n";
      croak "Don't understand data type $parse[0]\n"
	unless (exists $parser{$parse[0]});

      $parser{$parse[0]}->(@parse[1..$#parse]);
    }
  }
}

if (scalar @unimplementedObjects > 0) {
  print "Following objects are not implemented:\n";
  print join " ",@unimplementedObjects;
  exit;
}

#print STDERR join "=", keys %objecttranslations;

# Make header:
# (should perhaps roll pudac.h contents in?)
print qq[
#include <alsa/asoundlib.h>
#include <signal.h>
] if ($opts{o} =~ /alsa/i);
print qq[
#include <search.h>
#define _GNU_SOURCE
/* for getopt_long_only */
#include <sys/poll.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <string.h>
#include "pudac.h"

unsigned int sample_rate;
unsigned int block_size;

];

# Make prototypes:
print "void object$_ (int inletnumber, const struct PDdata * message);\n"
 for (sort {$a <=> $b } keys %objectlist);

# sort connects for each object (very important!)
# puredata sends things out of objects from rightmost outlet first
#  (which i store as "highest")
for (keys %objectlist) {
  if (exists $objectlist{$_}->{'AOCS'}) {
    my @t = split /\n/,$objectlist{$_}->{'AOCS'};
    @t = map { $_->[0] }
      sort { $b->[1] cmp $a->[1] }
	map { [$_, /,(.+)$/] }
	  @t;
    $objectlist{$_}->{'AOCS'} = join "\n",@t;
  }
}

# Generate code for objects:
for (sort {$b <=> $a } keys %objectlist) { # because send/receive go by *reverse object order*
  print "/* @{$objectlist{$_}->{'ARGS'}} */\n";
  if (exists $objectlist{$_}->{'CODE'}) {
    $objectlist{$_}->{'CODE'}
      ($_,
       %{$objectlist{$_}});
  } else {
    croak "Didn't have a code routine for $_: Only had ".join(", ",keys %{$objectlist{$_}});
  }
}

print qq`
snd_pcm_t * handle;

void capcc(int signo) {
  psignal(signo,"quitting because of signal");
  exit(1);
}

void alsaquit(void) {
  snd_pcm_close(handle);
}
` if ($opts{o} =~ /alsa/);

print qq`
int main ( void ) {
  struct PDdata DSPtracer={PDC_DSP_REQUEST,0,NULL,0,0,NULL,NULL};
  struct timeval now;
  int16_t * pcm_s16_out;
`;
print qq`
  snd_pcm_sframes_t frames;
` if ($opts{o} =~ /alsa/);
print qq`
  int err;
  int i;

  sample_rate = 48000;
  block_size = 256;
  DSPtracer.dsp_stream = calloc(block_size * 2,sizeof(float));
  pcm_s16_out = calloc(block_size * 2,sizeof(int16_t));
  if (DSPtracer.dsp_stream == NULL || pcm_s16_out == NULL) {
      error("Unable to allocate memory for root dac structure");
      exit(1);
  }

  hcreate(1000); // Maximum number of listening names.
  // It runs out of space if we use it up.
`;
print qq`
  atexit(alsaquit);
  for (i=1;i<32;i++)
    signal(i,capcc);
  // capture and quit on all signals

  // open libasound stuff
  // DSP:
  err=snd_pcm_open(&handle, "default", SND_PCM_STREAM_PLAYBACK, 0);
  if (err < 0) {
    error("playback open failure: \%s",snd_strerror(err));
    exit(1);
  }
  err=snd_pcm_set_params(handle, SND_PCM_FORMAT_S16, SND_PCM_ACCESS_RW_INTERLEAVED,
    /*channels:*/2, sample_rate, /*allow resampling:*/0, /*latency:*/10000);
  if (err < 0) {
    error("playback params failure: \%s ",snd_strerror(err));
    exit(1);
  }
` if ($opts{o} =~ /alsa/i);
print qq`
  $globalinit

  $loadbangs
  while ( 1 ) {
    gettimeofday(&now,NULL); /* we should tie this to the sample clock */
    $periteration
    memset(DSPtracer.dsp_stream,0,block_size*sizeof(float)*2);
    $dsptrace
`;
if ($opts{o} =~ /alsa/i) {
print qq`
    for (i=0;i<block_size*2;i++) {
      if (DSPtracer.dsp_stream[i]>=1) pcm_s16_out[i*2] = 32767;
      else if (DSPtracer.dsp_stream[i]<-1) pcm_s16_out[i*2] = -32768;
      else pcm_s16_out[i*2] = DSPtracer.dsp_stream[i]*32768;
      if (DSPtracer.dsp_stream[i+block_size]>=1) pcm_s16_out[i*2+1] = 32767;
      else if (DSPtracer.dsp_stream[i+block_size]<-1) pcm_s16_out[i*2+1] = -32768;
      else pcm_s16_out[i*2+1] = DSPtracer.dsp_stream[i+block_size]*32768;
    }
    frames = snd_pcm_writei(handle,pcm_s16_out,block_size);
    if (frames < 0) {
      error("error writing- \%s",frames,snd_strerror(frames));
      frames = snd_pcm_recover(handle,frames,0);
    }
    else if (frames < block_size)
      error("short write? (\%d < \%d",frames,block_size);
    if (frames < 0) {
      error("playback data failure: \%s",snd_strerror(frames));
      exit(1);
    }
    DSPtracer.sample_count += frames;
  }
}
` } else {
print qq`
    for (i=0;i<block_size;i++) {
      if (DSPtracer.dsp_stream[i]>=1) pcm_s16_out[i*2] = 32767;
      else if (DSPtracer.dsp_stream[i]<-1) pcm_s16_out[i*2] = -32768;
      else pcm_s16_out[i*2] = DSPtracer.dsp_stream[i]*32768;
      if (DSPtracer.dsp_stream[i+block_size]>=1) pcm_s16_out[i*2+1] = 32767;
      else if (DSPtracer.dsp_stream[i+block_size]<-1) pcm_s16_out[i*2+1] = -32768;
      else pcm_s16_out[i*2+1] = DSPtracer.dsp_stream[i+block_size]*32768;
    }
    write(1,pcm_s16_out,4*block_size);
    DSPtracer.sample_count += block_size;
  }
}
`;
}
